#!/usr/bin/env python3
"""
Password Policy Enforcer
Advanced password policy enforcement and management tool.
Parameters:
    - mode: Enforcement mode (check, enforce)
    - policy: Policy configuration file
    - output: Output file for results
Requires: passlib, bcrypt packages
"""
import argparse
import logging
from datetime import datetime
import json
import re
import hashlib
import getpass
import os
import sys
from pathlib import Path
import sqlite3
from passlib.hash import bcrypt
import string
import random

class PasswordPolicyEnforcer:
    def __init__(self, mode='check', policy_file=None):
        self.mode = mode.lower()
        self.policy_file = policy_file
        self.results = {}
        self.setup_logging()
        self.initialize_database()
        self.load_policy()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def initialize_database(self):
        """Initialize SQLite database for password history"""
        try:
            self.db_path = 'password_history.db'
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    CREATE TABLE IF NOT EXISTS password_history (
                        username TEXT,
                        password_hash TEXT,
                        timestamp TEXT,
                        PRIMARY KEY (username, password_hash)
                    )
                ''')
                
                cursor.execute('''
                    CREATE TABLE IF NOT EXISTS policy_violations (
                        username TEXT,
                        violation_type TEXT,
                        details TEXT,
                        timestamp TEXT
                    )
                ''')
                conn.commit()
        except Exception as e:
            self.logger.error(f"Error initializing database: {str(e)}")

    def load_policy(self):
        """Load password policy configuration"""
        if self.policy_file and os.path.exists(self.policy_file):
            try:
                with open(self.policy_file, 'r') as f:
                    self.policy = json.load(f)
            except Exception as e:
                self.logger.error(
                    f"Error loading policy file: {str(e)}"
                )
                self.policy = self.get_default_policy()
        else:
            self.policy = self.get_default_policy()

    def get_default_policy(self):
        """Get default password policy"""
        return {
            'min_length': 12,
            'max_length': 128,
            'require_uppercase': True,
            'require_lowercase': True,
            'require_numbers': True,
            'require_special': True,
            'special_chars': '!@#$%^&*()_+-=[]{}|;:,.<>?',
            'disallow_username': True,
            'disallow_common': True,
            'history_size': 5,
            'max_age_days': 90,
            'min_age_days': 1,
            'complexity_score': 3,
            'banned_words': [
                'password', 'admin', 'root', 'user',
                '123456', 'qwerty', 'letmein'
            ],
            'pattern_blacklist': [
                r'(.)\1{2,}',  # Repeated characters
                r'12345',      # Sequential numbers
                r'qwerty',     # Keyboard patterns
                r'asdfgh'
            ]
        }

    def check_password_strength(self, username, password):
        """Check password strength against policy"""
        violations = []
        
        # Check length
        if len(password) < self.policy['min_length']:
            violations.append(
                f"Password too short (minimum {self.policy['min_length']} characters)"
            )
        if len(password) > self.policy['max_length']:
            violations.append(
                f"Password too long (maximum {self.policy['max_length']} characters)"
            )

        # Check character requirements
        if (self.policy['require_uppercase'] and
            not any(c.isupper() for c in password)):
            violations.append("Missing uppercase letter")
        
        if (self.policy['require_lowercase'] and
            not any(c.islower() for c in password)):
            violations.append("Missing lowercase letter")
        
        if (self.policy['require_numbers'] and
            not any(c.isdigit() for c in password)):
            violations.append("Missing number")
        
        if (self.policy['require_special'] and
            not any(c in self.policy['special_chars'] for c in password)):
            violations.append("Missing special character")

        # Check username
        if (self.policy['disallow_username'] and
            username.lower() in password.lower()):
            violations.append("Password contains username")

        # Check common passwords
        if self.policy['disallow_common']:
            if self._is_common_password(password):
                violations.append("Password is too common")

        # Check banned words
        for word in self.policy['banned_words']:
            if word.lower() in password.lower():
                violations.append(f"Password contains banned word: {word}")

        # Check patterns
        for pattern in self.policy['pattern_blacklist']:
            if re.search(pattern, password):
                violations.append("Password contains forbidden pattern")

        # Check password history
        if self._is_password_reused(username, password):
            violations.append(
                f"Password was used in the past {self.policy['history_size']} passwords"
            )

        # Calculate complexity score
        score = self._calculate_complexity_score(password)
        if score < self.policy['complexity_score']:
            violations.append(
                f"Password complexity score too low ({score} < {self.policy['complexity_score']})"
            )

        return violations

    def _is_common_password(self, password):
        """Check if password is in common password list"""
        common_passwords = set([
            'password123', 'admin123', '123456789',
            'qwerty123', 'letmein123', 'welcome123'
        ])
        return password.lower() in common_passwords

    def _is_password_reused(self, username, password):
        """Check if password exists in history"""
        try:
            password_hash = bcrypt.hash(password)
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    SELECT COUNT(*) FROM password_history
                    WHERE username = ?
                    ORDER BY timestamp DESC
                    LIMIT ?
                ''', (username, self.policy['history_size']))
                
                count = cursor.fetchone()[0]
                return count > 0 and any(
                    bcrypt.verify(
                        password,
                        stored_hash[0]
                    ) for stored_hash in cursor.fetchall()
                )
        except Exception as e:
            self.logger.error(
                f"Error checking password history: {str(e)}"
            )
            return False

    def _calculate_complexity_score(self, password):
        """Calculate password complexity score"""
        score = 0
        
        # Length score
        score += len(password) // 8

        # Character variety score
        if any(c.isupper() for c in password):
            score += 1
        if any(c.islower() for c in password):
            score += 1
        if any(c.isdigit() for c in password):
            score += 1
        if any(c in self.policy['special_chars'] for c in password):
            score += 1

        # Pattern score
        if not any(
            re.search(pattern, password)
            for pattern in self.policy['pattern_blacklist']
        ):
            score += 1

        # Entropy score
        entropy = self._calculate_entropy(password)
        score += entropy // 20

        return score

    def _calculate_entropy(self, password):
        """Calculate password entropy"""
        char_sets = {
            'uppercase': string.ascii_uppercase,
            'lowercase': string.ascii_lowercase,
            'digits': string.digits,
            'special': self.policy['special_chars']
        }
        
        pool_size = 0
        for char_set in char_sets.values():
            if any(c in char_set for c in password):
                pool_size += len(char_set)

        return len(password) * (pool_size.bit_length())

    def store_password(self, username, password):
        """Store password hash in history"""
        try:
            password_hash = bcrypt.hash(password)
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    INSERT INTO password_history
                    (username, password_hash, timestamp)
                    VALUES (?, ?, ?)
                ''', (
                    username,
                    password_hash,
                    datetime.now().isoformat()
                ))
                conn.commit()
            return True
        except Exception as e:
            self.logger.error(
                f"Error storing password: {str(e)}"
            )
            return False

    def log_violation(self, username, violation_type, details):
        """Log policy violation"""
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    INSERT INTO policy_violations
                    (username, violation_type, details, timestamp)
                    VALUES (?, ?, ?, ?)
                ''', (
                    username,
                    violation_type,
                    details,
                    datetime.now().isoformat()
                ))
                conn.commit()
        except Exception as e:
            self.logger.error(
                f"Error logging violation: {str(e)}"
            )

    def generate_password(self):
        """Generate policy-compliant password"""
        while True:
            # Initialize character pools
            lowercase = string.ascii_lowercase
            uppercase = string.ascii_uppercase
            digits = string.digits
            special = self.policy['special_chars']

            # Calculate minimum length for each required character type
            min_length = max(
                self.policy['min_length'],
                sum([
                    self.policy['require_lowercase'],
                    self.policy['require_uppercase'],
                    self.policy['require_numbers'],
                    self.policy['require_special']
                ])
            )

            # Generate password
            password = []
            
            # Add required characters
            if self.policy['require_lowercase']:
                password.append(random.choice(lowercase))
            if self.policy['require_uppercase']:
                password.append(random.choice(uppercase))
            if self.policy['require_numbers']:
                password.append(random.choice(digits))
            if self.policy['require_special']:
                password.append(random.choice(special))

            # Fill remaining length with random characters
            all_chars = ''
            if self.policy['require_lowercase']:
                all_chars += lowercase
            if self.policy['require_uppercase']:
                all_chars += uppercase
            if self.policy['require_numbers']:
                all_chars += digits
            if self.policy['require_special']:
                all_chars += special

            while len(password) < min_length:
                password.append(random.choice(all_chars))

            # Shuffle password
            random.shuffle(password)
            password = ''.join(password)

            # Verify password meets all requirements
            if not self.check_password_strength('', password):
                return password

    def check_password_age(self, username):
        """Check password age against policy"""
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    SELECT timestamp FROM password_history
                    WHERE username = ?
                    ORDER BY timestamp DESC
                    LIMIT 1
                ''', (username,))
                
                result = cursor.fetchone()
                if result:
                    last_change = datetime.fromisoformat(result[0])
                    age_days = (
                        datetime.now() - last_change
                    ).days
                    
                    if age_days > self.policy['max_age_days']:
                        return {
                            'expired': True,
                            'days_old': age_days,
                            'max_age': self.policy['max_age_days']
                        }
                    elif age_days < self.policy['min_age_days']:
                        return {
                            'too_new': True,
                            'days_old': age_days,
                            'min_age': self.policy['min_age_days']
                        }
                    
                    return {
                        'expired': False,
                        'too_new': False,
                        'days_old': age_days
                    }
                    
                return None
        except Exception as e:
            self.logger.error(
                f"Error checking password age: {str(e)}"
            )
            return None

    def generate_report(self):
        """Generate password policy report"""
        try:
            report = {
                'timestamp': datetime.now().isoformat(),
                'policy': self.policy,
                'statistics': {
                    'total_users': 0,
                    'compliant_users': 0,
                    'expired_passwords': 0,
                    'violations': defaultdict(int)
                }
            }

            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                
                # Get user statistics
                cursor.execute(
                    'SELECT COUNT(DISTINCT username) FROM password_history'
                )
                report['statistics']['total_users'] = cursor.fetchone()[0]

                # Get violation statistics
                cursor.execute('''
                    SELECT violation_type, COUNT(*)
                    FROM policy_violations
                    GROUP BY violation_type
                ''')
                for violation_type, count in cursor.fetchall():
                    report['statistics']['violations'][violation_type] = count

                # Check password ages
                cursor.execute(
                    'SELECT DISTINCT username FROM password_history'
                )
                for (username,) in cursor.fetchall():
                    age_check = self.check_password_age(username)
                    if age_check:
                        if age_check['expired']:
                            report['statistics']['expired_passwords'] += 1
                        elif not age_check['too_new']:
                            report['statistics']['compliant_users'] += 1

            return report
        except Exception as e:
            self.logger.error(f"Error generating report: {str(e)}")
            return None

    def save_report(self, report, output_file):
        """Save policy report to file"""
        try:
            with open(output_file, 'w') as f:
                json.dump(report, f, indent=4)
            self.logger.info(f"Report saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving report: {str(e)}")

    def print_summary(self, report):
        """Print policy enforcement summary"""
        self.logger.info("\nPassword Policy Summary:")
        
        self.logger.info("\nPolicy Requirements:")
        self.logger.info(
            f"Minimum Length: {self.policy['min_length']}"
        )
        self.logger.info(
            f"Maximum Length: {self.policy['max_length']}"
        )
        self.logger.info(
            f"Require Uppercase: {self.policy['require_uppercase']}"
        )
        self.logger.info(
            f"Require Lowercase: {self.policy['require_lowercase']}"
        )
        self.logger.info(
            f"Require Numbers: {self.policy['require_numbers']}"
        )
        self.logger.info(
            f"Require Special: {self.policy['require_special']}"
        )

        if 'statistics' in report:
            stats = report['statistics']
            self.logger.info("\nEnforcement Statistics:")
            self.logger.info(f"Total Users: {stats['total_users']}")
            self.logger.info(
                f"Compliant Users: {stats['compliant_users']}"
            )
            self.logger.info(
                f"Expired Passwords: {stats['expired_passwords']}"
            )
            
            if stats['violations']:
                self.logger.info("\nViolations:")
                for violation, count in stats['violations'].items():
                    self.logger.info(
                        f"{violation}: {count}"
                    )

    def enforce_policy(self, username, password=None):
        """Enforce password policy for user"""
        if self.mode != 'enforce':
            return False

        try:
            # Check existing password age if no new password provided
            if not password:
                age_check = self.check_password_age(username)
                if age_check:
                    if age_check['expired']:
                        self.log_violation(
                            username,
                            'expired_password',
                            f"Password expired ({age_check['days_old']} days old)"
                        )
                        return False
                    elif age_check['too_new']:
                        self.log_violation(
                            username,
                            'password_too_new',
                            f"Password too new ({age_check['days_old']} days old)"
                        )
                        return False
                return True

            # Check new password strength
            violations = self.check_password_strength(username, password)
            if violations:
                for violation in violations:
                    self.log_violation(
                        username,
                        'policy_violation',
                        violation
                    )
                return False

            # Store new password
            return self.store_password(username, password)

        except Exception as e:
            self.logger.error(f"Error enforcing policy: {str(e)}")
            return False

    def start_enforcement(self, output_file=None):
        """Start password policy enforcement"""
        self.logger.info("Starting password policy enforcement")
        self.logger.info(f"Mode: {self.mode}")
        
        try:
            # Generate and save report
            report = self.generate_report()
            if report and output_file:
                self.save_report(report, output_file)
            if report:
                self.print_summary(report)
                
        except KeyboardInterrupt:
            self.logger.info("Enforcement interrupted by user")
        except Exception as e:
            self.logger.error(f"Error during enforcement: {str(e)}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Password Policy Enforcer"
    )
    parser.add_argument(
        "--mode",
        choices=['check', 'enforce'],
        default='check',
        help="Enforcement mode"
    )
    parser.add_argument(
        "--policy",
        help="Policy configuration file"
    )
    parser.add_argument(
        "--output",
        help="Output file for policy report"
    )
    args = parser.parse_args()

    enforcer = PasswordPolicyEnforcer(args.mode, args.policy)
    enforcer.start_enforcement(args.output)
